/* Selection.java
 *
 * created: Fri Nov  6 1998
 *
 * This file is part of Artemis
 *
 * Copyright (C) 1998,1999  Genome Research Limited
 *
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 *
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 *
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/diana/Selection.java,v 1.23 1999/10/13 15:42:00 kmr Exp $
 */

package diana;

import diana.sequence.*;

import java.io.*;
import java.util.Vector;
import java.awt.datatransfer.*;

/**
 *  Objects of this class hold the features/feature segments or entries that
 *  the user has selected.
 *
 *  @author Kim Rutherford
 *  @version $Id: Selection.java,v 1.23 1999/10/13 15:42:00 kmr Exp $
 *
 **/

public class Selection
    implements FeatureChangeListener, EntryChangeListener, Transferable,
               ClipboardOwner {
  /**
   *  Create a new Selection object.
   *  @param clipboard The system clipboard reference.  This may be null if
   *    this Selection should not set the system clipboard.
   **/
  public Selection (final Clipboard clipboard) {
    this.clipboard = clipboard;
  }

  private final DataFlavor flavors [] = {
    DataFlavor.stringFlavor, DataFlavor.plainTextFlavor
  };

  /**
   *  Returns the array of the data flavors that this object can provide.
   **/
  public synchronized DataFlavor[] getTransferDataFlavors () {
    return flavors;
  }

  /**
   *  Returns whether the requested flavor is supported by this object.
   *  @param flavor the requested flavor for the data
   **/
  public boolean isDataFlavorSupported (final DataFlavor flavor) {
    return
      flavor.equals (DataFlavor.stringFlavor) ||
      flavor.equals (DataFlavor.plainTextFlavor);
  }

  /**
   *  If the data was requested in the "java.lang.String" flavor, return the
   *  String representing the selection, else throw an
   *  UnsupportedFlavorException.
   *  @param flavor the requested flavor for the data */
  public synchronized Object getTransferData (final DataFlavor flavor)
      throws UnsupportedFlavorException, IOException {
    if (flavor.equals (DataFlavor.stringFlavor)) {
      return getSelectionText ();
    } else if (flavor.equals (DataFlavor.plainTextFlavor)) {
      return new StringReader (getSelectionText ());
    } else {
      throw new UnsupportedFlavorException (flavor);
    }
  }

  /**
   *  Implementation of the ClipboardOwner interface.
   **/
  public void lostOwnership(final Clipboard clipboard,
                            final Transferable contents) {

  }

  /**
   *  Returns this Selection in a String.
   **/
  private String getSelectionText () {
    final FeatureVector selection_features = getAllFeatures ();

    final StringBuffer buffer = new StringBuffer ();

    final MarkerRange marker_range = getMarkerRange ();

    if (marker_range != null) {
      final String selection_bases = Strand.markerRangeBases (marker_range);

      buffer.append (selection_bases);
    }

    for (int i = 0 ; i < selection_features.size () && i < 50 ; ++i) {
      buffer.append (selection_features.elementAt (i).toString ());
    }

    return buffer.toString ();
  }

  /**
   *  Implementation of the FeatureChangeListener interface.  We need to
   *  listen to feature change events so a SelectionChangeEvent can be sent if
   *  a selected feature is changed.
   *  @param event The change event.
   **/
  public void featureChanged (FeatureChangeEvent event) {
//       System.out.println ("Selection: feature change event type: " +
//                           event.getType ());

    if (getAllFeatures ().contains (event.getFeature ())) {
      changeSelection ();
    }
  }

  /**
   *  Implementation of the EntryChangeListener interface.  We listen to
   *  EntryChange events so that we remove feature from the selection when
   *  they are removed from the Entry.
   **/
  public void entryChanged (final EntryChangeEvent event) {
//      System.out.println ("in Selection entry change event type: " +
//                          event.getType () + " for feature " +
//                          event.getFeature ());

    switch (event.getType ()) {
    case EntryChangeEvent.FEATURE_DELETED:

      features.remove (event.getFeature ());

      // this will fire a SelectionChangeEvent
      removeSegmentsOf (event.getFeature ());
      break;
    }
  }

  /**
   *  Adds the specified event listener to receive selection change events
   *  from this object.
   *  @param l the event change listener.
   **/
  public void addSelectionChangeListener (final SelectionChangeListener l) {
    selection_listener_list.addElement (l);
  }

  /**
   *  Removes the specified event listener so that it no longer receives
   *  selection change events from this object.
   *  @param l the event change listener.
   **/
  public void removeSelectionChangeListener (final SelectionChangeListener l) {
    selection_listener_list.removeElement (l);
  }

  /**
   *  Broadcast an event to notify SelectionChange event listeners that the
   *  selection has changed.
   **/
  private void changeSelection () {
    final SelectionChangeEvent event =
      new SelectionChangeEvent (this);

    fireAction (selection_listener_list, event);

    if (clipboard != null) {
      clipboard.setContents (this, this);
    }
  }

  /**
   *  Add a Feature to the selection if it isn't there already.
   **/
  public void add (final Feature feature) {
    if (addWithoutEvent (feature)) {
      changeSelection ();
    }
  }

  /**
   *  Add the given Feature objects to the selection, and then send an
   *  appropriate event to the SelectionChangeEvent listeners.
   **/
  public void add (final FeatureVector features) {
    for (int i = 0 ; i < features.size () ; ++i) {
      addWithoutEvent (features.elementAt (i));
    }
    changeSelection ();
  }

  /**
   *  Add a Feature to the selection if it isn't there already, but don't send
   *  an event to the SelectionChangeEvent listeners.
   *  @return true if and only if the segment with not already in the
   *    selection.
   **/
  private boolean addWithoutEvent (final Feature feature) {
    if (features.contains (feature)) {
      return false;
      // do nothing
    } else {
      features.add (feature);
      return true;
    }
  }

  /**
   *  Add a FeatureSegment to the selection if it isn't there already, and
   *  send an event to the SelectionChangeEvent listeners..
   **/
  public void add (final FeatureSegment segment) {
    if (addWithoutEvent (segment)) {
      changeSelection ();
    }
  }

  /**
   *  Add a FeatureSegment to the selection if it isn't there already, but
   *  don't send an event to the SelectionChangeEvent listeners.
   *  @return true if and only if the segment with not already in the
   *    selection.
   **/
  private boolean addWithoutEvent (final FeatureSegment segment) {
    if (segments.contains (segment)) {
      return false;
      // do nothing
    } else {
      segments.addElement (segment);
      return true;
    }
  }

  /**
   *  Remove all the objects from the selection, add the given Feature object
   *  to the selection, and then send an appropriate event to the
   *  SelectionChangeEvent listeners.
   **/
  public void set (final Feature feature) {
    clearWithoutEvent ();
    addWithoutEvent (feature);
    changeSelection ();
  }

  /**
   *  Remove all the objects from the selection, add the given FeatureSegment
   *  object to the selection, and then send an appropriate event to the
   *  SelectionChangeEvent listeners.
   **/
  public void set (final FeatureSegment feature_segment) {
    clearWithoutEvent ();
    addWithoutEvent (feature_segment);
    changeSelection ();
  }

  /**
   *  Remove all the objects from the selection, add the given Feature objects
   *  to the selection, and then send an appropriate event to the
   *  SelectionChangeEvent listeners.
   **/
  public void set (final FeatureVector features) {
    clearWithoutEvent ();
    for (int i = 0 ; i < features.size () ; ++i) {
      addWithoutEvent (features.elementAt (i));
    }
    changeSelection ();
  }

  /**
   *  Set the MarkerRange object that this Selection holds and remove all
   *  other objects.
   **/
  public void setMarkerRange (final MarkerRange marker_range) {
    if (this.marker_range != marker_range) {
      clearWithoutEvent ();
      this.marker_range = marker_range;
      changeSelection ();
    }
  }

  /**
   *  Return true if and only if there is nothing in the selection.
   **/
  public boolean isEmpty () {
    if (features.size () == 0 &&
        segments.size () == 0 &&
        marker_range == null) {
      return true;
    } else {
      return false;
    }
  }

  /**
   *  Remove all the objects from the selection, and send an event to the
   *  SelectionChangeEvent listeners.
   **/
  public void clear () {
    if (!isEmpty ()) {
      clearWithoutEvent ();
      changeSelection ();
    }
  }

  /**
   *  Remove all the objects from the selection, but don't send an event to
   *  the SelectionChangeEvent listeners.
   **/
  private void clearWithoutEvent () {
    features.removeAllElements ();
    segments.removeAllElements ();
    marker_range = null;
  }

  /**
   *  Remove a Feature from the selection.
   **/
  public void remove (final Feature feature) {
    if (features.remove (feature)) {
      changeSelection ();
    }
  }

  /**
   *  Remove the FeatureSegments of the given Feature from the selection.
   **/
  public void removeSegmentsOf (final Feature feature) {
    final FeatureSegmentVector segments = feature.getSegments ();

    for (int segment_index = 0 ;
         segment_index < segments.size () ;
         ++segment_index) {
      final FeatureSegment this_segment = segments.elementAt (segment_index);

      remove (this_segment);
    }
    changeSelection ();
  }

  /**
   *  Remove a FeatureSegment from the selection.
   **/
  public void remove (final FeatureSegment segment) {
    if (segments.removeElement (segment)) {
      changeSelection ();
    }
  }

  /**
   *  Return true if this selection contains the given Feature.
   **/
  public boolean contains (final Feature feature) {
    if (getSelectedFeatures ().contains (feature)) {
      return true;
    }

    for (int i = 0 ; i < segments.size () ; ++i) {
      final FeatureSegment this_segment = segments.elementAt (i);
      final Feature this_feature = this_segment.getFeature ();

      if (feature == this_feature) {
        return true;
      }
    }

    return false;
  }

  /**
   *  Return true if this selection contains the given Feature.
   **/
  public boolean contains (final FeatureSegment feature_segment) {
    if (getAllSegments ().indexOf (feature_segment) == -1) {
      return false;
    } else {
      return true;
    }
  }

  /**
   *  Return the first base of the MarkerRange (if it is set), or the first
   *  base of the first selected feature (if there are any selected features),
   *  or the first selected FeatureSegment (if there are any selected
   *  FeatureSegment objects), otherwise null.
   **/
  public Marker getFirstBaseOfSelection () {
    if (getSelectedFeatures ().size () > 0) {
      return getSelectedFeatures ().elementAt (0).getFirstBaseMarker ();
    } else {
      if (getSelectedSegments ().size () > 0) {
        return getSelectedSegments ().elementAt (0).getStart ();
      } else {
        if (getMarkerRange () == null) {
          return null;
        } else {
          return getMarkerRange ().getStart ();
        }
      }
    }
  }

  /**
   *  Return the last base of the MarkerRange (if it is set), or the last base
   *  of the last selected feature (if there are any selected features), or
   *  the last selected FeatureSegment (if there are any selected
   *  FeatureSegment objects), otherwise null.
   **/
  public Marker getLastBaseOfSelection () {
    if (getSelectedFeatures ().size () > 0) {
      return getSelectedFeatures ().lastElement ().getLastBaseMarker ();
    } else {
      if (getSelectedSegments ().size () > 0) {
        return getSelectedSegments ().lastElement ().getEnd ();
      } else {
        if (getMarkerRange () == null) {
          return null;
        } else {
          return getMarkerRange ().getEnd ();
        }
      }
    }
  }

  /**
   *  Return the bases of the selected features or the selected marker range.
   **/
  public String getSelectedBases () {
    if (getMarkerRange () == null) {
      final StringBuffer buffer = new StringBuffer ();

      final FeatureVector all_features = getAllFeatures ();

      for (int i = 0 ; i < all_features.size () ; ++i) {
        final Feature this_feature = all_features.elementAt (i);
        buffer.append (this_feature.getBases ());
      }

      return buffer.toString ();
    } else {
      return Strand.markerRangeBases (getMarkerRange ());
    }
  }

  /**
   *  Return the object that was set with the setMarkerRange (MarkerRange)
   *  method.
   **/
  public MarkerRange getMarkerRange () {
    return marker_range;
  }

  /**
   *  Return a vector of the Feature objects of this selection.
   **/
  public FeatureVector getSelectedFeatures () {
    return features;
  }

  /**
   *  Return a vector of the FeatureSegment objects of this selection.
   **/
  public FeatureSegmentVector getSelectedSegments () {
    return segments;
  }

  /**
   *  Return a vector of all the FeatureSegment objects of this selection.
   *  This will include the FeatureSegment objects that are owned by Feature
   *  objects in the selection.
   **/
  public FeatureSegmentVector getAllSegments () {
    final FeatureSegmentVector return_segments = new FeatureSegmentVector ();

    for (int i = 0 ; i < features.size () ; ++i) {
      final Feature selection_feature = features.elementAt (i);
      final FeatureSegmentVector segments = selection_feature.getSegments ();

      for (int segment_index = 0 ;
           segment_index < segments.size () ;
           ++segment_index) {
        final FeatureSegment this_segment = segments.elementAt (segment_index);

        if (return_segments.indexOf (this_segment) == -1) {
          return_segments.addElement (this_segment);
        }
      }
    }

    for (int i = 0 ; i < segments.size () ; ++i) {
      final FeatureSegment this_segment = segments.elementAt (i);

      if (return_segments.indexOf (this_segment) == -1) {
        return_segments.addElement (this_segment);
      }
    }

    return return_segments;
  }

  /**
   *  Return a vector of all the Feature objects of this selection.  This will
   *  include the Feature objects that own the FeatureSegment objects in the
   *  selection.
   **/
  public FeatureVector getAllFeatures () {
    final FeatureVector return_features = (FeatureVector) features.clone ();

    for (int i = 0 ; i < segments.size () ; ++i) {
      final FeatureSegment this_segment = segments.elementAt (i);
      final Feature this_feature = this_segment.getFeature ();

      if (!return_features.contains (this_feature)) {
        return_features.add (this_feature);
      }
    }

    return return_features;
  }

  /**
   *  Send an event to those object listening for it.
   *  @param listeners A Vector of the objects that the event should be sent
   *    to.
   *  @param event The event to send
   **/
  private void fireAction (final Vector listeners, final ChangeEvent event) {
    final Vector targets;
    // copied from a book - synchronising the whole method might cause a
    // deadlock
    synchronized (this) {
      targets = (Vector) listeners.clone ();
    }

    for ( int i = 0 ; i < targets.size () ; ++i ) {
      ChangeListener target = (ChangeListener) targets.elementAt (i);

      if (event instanceof SelectionChangeEvent) {
        final SelectionChangeListener selection_change_listener =
          (SelectionChangeListener) target;
        selection_change_listener.selectionChanged ((SelectionChangeEvent) event);
      } else {
        throw new Error ("Selection.fireAction () - unknown event");
      }
    }
  }

  /**
   *  A Vector containing the Feature objects that this selection currently
   *  holds.
   **/
  private FeatureVector features = new FeatureVector ();

  /**
   *  A Vector containing the FeatureSegment objects that this selection
   *  currently holds.
   **/
  private FeatureSegmentVector segments = new FeatureSegmentVector ();

  /**
   *  Each Selection object can hold one MarkerRange.
   **/
  private MarkerRange marker_range = null;

  /**
   *  A vector of those objects listening for selection change events.
   **/
  final private Vector selection_listener_list = new Vector ();

  /**
   *  The system clipboard as passed to the constructor.
   **/
  final private Clipboard clipboard;
}
