/* PublicDBStreamFeature.java
 *
 * created: Tue Sep 14 1999
 *
 * This file is part of Artemis
 * 
 * Copyright (C) 1999  Genome Research Limited
 * 
 * This program is free software; you can redistribute it and/or
 * modify it under the terms of the GNU General Public License
 * as published by the Free Software Foundation; either version 2
 * of the License, or (at your option) any later version.
 * 
 * This program is distributed in the hope that it will be useful,
 * but WITHOUT ANY WARRANTY; without even the implied warranty of
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
 * GNU General Public License for more details.
 * 
 * You should have received a copy of the GNU General Public License
 * along with this program; if not, write to the Free Software
 * Foundation, Inc., 59 Temple Place - Suite 330, Boston, MA  02111-1307, USA.
 *
 * $Header: /nfs/disk222/yeastpub/Repository/powmap/uk/ac/sanger/pathogens/embl/PublicDBStreamFeature.java,v 1.2 1999/09/20 08:36:10 kmr Exp $
 */

package uk.ac.sanger.pathogens.embl;

import uk.ac.sanger.pathogens.*;
import java.io.*;

/**
 *  This is an implementation of Feature that can read and write itself to a
 *  EMBL or GENBANK stream.
 *
 *  @author Kim Rutherford
 *  @version $Id: PublicDBStreamFeature.java,v 1.2 1999/09/20 08:36:10 kmr Exp $
 **/

abstract class PublicDBStreamFeature
    extends LineGroup
    implements StreamFeature, ComparableFeature {
  /**
   *  Create a new PublicDBStreamFeature object.  The feature should be added
   *  to an Entry (with Entry.add ()).
   *  @param key The new feature key
   *  @param location The Location object for the new feature
   *  @param qualifiers The qualifiers for the new feature
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    the given Qualifier.
   **/
  public PublicDBStreamFeature (Key key,
                                Location location,
                                QualifierVector qualifiers)
      throws InvalidRelationException {
    try {
      setKey (key);
      setLocation (location);
      setQualifiers (qualifiers);
    } catch (ReadOnlyException e) {
      System.err.println ("internal error - unexpected exception: " + e);
    }
  }

  /**
   *  Create a new PublicDBStreamFeature with the same key, location and
   *  qualifiers as the given feature.  The feature should be added to an
   *  Entry (with Entry.add ()).
   *  @param feature The feature to copy.
   **/
  public PublicDBStreamFeature (final Feature feature) {
    try {
      setEntry (null);
      setKey (feature.getKey ());
      setLocation (feature.getLocation ());
      setQualifiers (feature.getQualifiers ());
    } catch (InvalidRelationException e) {
      throw new Error ("unexpected exception while creating a Feature " +
                       "object - " + e);
    } catch (ReadOnlyException e) {
      System.err.println ("internal error - unexpected exception: " + e);
    }
  }

  /**
   *  This is incremented each time a constructor is called.
   **/
  private static long id_counter = 0;

  /**
   *  Set the key, location and qualifiers of this object to be the same as
   *  the given Feature.
   *  @param key The new feature key
   *  @param location The Location object for the new feature
   *  @param qualifiers The qualifiers for the new feature
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    the given Qualifier.
   **/
  public void set (final Feature feature)
      throws InvalidRelationException, ReadOnlyException {
    set (feature.getKey (), feature.getLocation (), feature.getQualifiers ());
  }
  
  /**
   *  Set the value of this object.
   *  @param key The new feature key
   *  @param location The Location object for the new feature
   *  @param qualifiers The qualifiers for the new feature
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    the given Qualifier.
   **/
  public void set (Key key,
                   Location location,
                   QualifierVector qualifiers)
      throws InvalidRelationException, ReadOnlyException {
    // save the old contents of the feature so they cannot be reset if there
    // is an exception
    for (int i = 0 ; i < qualifiers.size () ; ++i) {
      final String this_name = qualifiers.elementAt (i).getName ();
      
      if (!Qualifier.isPossibleQualifier (key, this_name)) {
        final String message =
          key + " cannot have /" + qualifiers.elementAt (i).getName () +
          " as a qualifier";
        throw new InvalidRelationException (message);
      }
    }

    this.key = key;
    this.qualifiers = qualifiers;
    setLocation (location);
  }

  /**
   *  Set the value of this object.
   *  @param entry The Entry that now owns this Feature.
   *  @param key The new feature key
   *  @param location The Location object for the new feature
   *  @param qualifiers The qualifiers for the new feature
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    the given Qualifier.
   **/
  private void set (DocumentEntry entry,
                    Key key,
                    Location location,
                    QualifierVector qualifiers)
      throws InvalidRelationException, ReadOnlyException {
    setEntry (entry);
    setKey (key);

    // throw away the cache
    first_base = -1;
    last_base = -1;    
    this.location = location;

    setQualifiers (qualifiers);
  }

  /**
   *  Set the owning Entry of this Feature.  Other objects should call
   *  setDocumentEntry () to change the owner.
   *  @param entry The Entry that now owns this Feature.
   **/
  private void setEntry (DocumentEntry entry) {
    this.entry = entry;
  }

  /**
   *  Set the owning DocumentEntry of this Feature.
   *  @param entry The Entry that now owns this Feature.
   **/
  void setDocumentEntry (DocumentEntry entry) {
    setEntry (entry);
  }

  /**
   *  Set the key field of this object.
   *  @param key The new feature key
   **/
  public void setKey (Key key)
      throws InvalidRelationException {
    for (int i = 0 ; i < qualifiers.size () ; ++i) {
      final String this_name = qualifiers.elementAt (i).getName ();
      
      if (!Qualifier.isPossibleQualifier (key, this_name)) {
        final String message =
          key + " cannot have /" + qualifiers.elementAt (i).getName () +
          " as a qualifier";
        throw new InvalidRelationException (message);
      }
    }
    this.key = key;
  }

  /**
   *  Set the location of this object.
   *  @param location The Location object for the new feature
   **/
  public void setLocation (Location location)
      throws ReadOnlyException {
    // save the Entry because the call to remove () will set it to null
    final Entry saved_entry = getEntry ();

    if (saved_entry != null) {
      // remove and then add the Feature because changing the Location may
      // change the position of the Feature in the FeatureTable
      saved_entry.remove (this);
    }

    this.location = location;

    // throw away the cache
    first_base = -1;
    last_base = -1;

    if (saved_entry != null) {
      saved_entry.add (this);
    }
  }

  /**
   *  Set the qualifiers of this object.
   *  @param qualifiers The qualifiers for the new feature
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    the given Qualifier.
   **/
  public void setQualifiers (QualifierVector qualifiers)
      throws InvalidRelationException {

    if (qualifiers == null) {
      qualifiers = new QualifierVector ();
    } else {
      for (int i = 0 ; i < qualifiers.size () ; ++i) {

        final String this_name = qualifiers.elementAt (i).getName ();

        if (!Qualifier.isPossibleQualifier (getKey (), this_name)) {
          final String message =
            getKey () + " cannot have /" +
            qualifiers.elementAt (i).getName () + " as a qualifier";
          throw new InvalidRelationException (message);
        }
      }
      this.qualifiers = qualifiers.copy ();
    }
  }

  /**
   *  Add the given Qualifier to this Feature.  If this Feature contains a
   *  Qualifier with the same name as the new Qualifier it will be replaced.
   *  @param qualifier The new qualifier to add.
   *  @exception InvalidRelationException Throw if this Feature cannot contain
   *    the given Qualifier.
   **/
  public void setQualifier (Qualifier qualifier)
      throws InvalidRelationException {
    getQualifiers ().setQualifier (qualifier);
  }

  /**
   *  Remove the Qualifier with the given name.  If there is no Qualifier with
   *  that name then return immediately.
   *  @param name The qualifier name to look for.
   **/
  public void removeQualiferByName (final String name)
      throws ReadOnlyException {
    getQualifiers ().removeQualiferByName (name);
  }
  
  /**
   *  Add the values from the given qualifier to the Qualifier object with the
   *  same name in this Feature or otherwise add a copy of the argument.
   *  @param qualifier This object contains name and values to add.
   *  @return The Qualifier that was changed or created.
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    the given Qualifier.
   **/
  public Qualifier addQualifierValues (Qualifier qualifier)
      throws InvalidRelationException {
    return getQualifiers ().addQualifierValues (qualifier);
  }

  /**
   *  Return the unique identifier of this PublicDBStreamFeature.
   **/
  public long getNumericID () {
    return id;
  }

  /**
   *  Return the Key of this Feature, as passed to the constructor.
   **/
  public Key getKey () {
    return key;
  }

  /**
   *  Return the Location of this Feature, as passed to the constructor.
   **/
  public Location getLocation () {
    return location;
  }

  /**
   *  Return the first base of this feature.
   **/
  public int getFirstBase () {
    if (first_base == -1) {
      first_base = getLocation ().getFirstBase ();
    }
    return first_base;
  }

  /**
   *  Return the last base of this feature.
   **/
  public int getLastBase () {
    if (last_base == -1) {
      last_base = getLocation ().getLastBase ();
    }
    return last_base;
  }

  /**
   *  Return a vector containing the qualifiers for this feature.  This method
   *  does not return a copy of the qualifier vector so changing the vector
   *  will change the feature.
   **/
  public QualifierVector getQualifiers () {
    return qualifiers;
  }

  /**
   *  Return the Qualifier in this Feature with the given name or null if
   *  there no such Qualifier.
   **/
  public Qualifier getQualifierByName (String name) {
    return getQualifiers ().getQualifierByName (name);
  }

  /**
   *  Return the Entry object that contains this Feature.
   **/
  public Entry getEntry () {
    return entry;
  }

  /**
   *  Return the DocumentEntry that conatins this PublicDBStreamFeature.  A
   *  PublicDBStreamFeature will always be owned by a DocumentEntry.
   **/
  private DocumentEntry getDocumentEntry () {
    return entry;
  }

  /**
   *  Return the reference of a new copy of this Feature.
   **/
  public abstract Feature copy ();

  /**
   *  Move this Feature to the given DocumentEntry
   **/
  void moveToDocumentEntry (final DocumentEntry entry)
      throws ReadOnlyException {
    if (getEntry () == entry) {
      // there is nothing to do
      return;
    }

    getEntry ().remove (this);

    entry.add (this);

    setEntry (entry);
  }

  /**
   *  Write this Feature to the given stream.  The output will be in EMBL
   *  format (but see the diana_extensions argument).
   *  @param writer The stream to write to.
   *  @param diana_extensions If true then the any diana additions to the embl
   *    file format will be included in the output, otherwise they will be
   *    removed.
   *  @exception IOException thrown if there is an io problem while writing
   *    the Feature.
   **/
  public void writeToStream (final Writer writer,
                             final boolean diana_extensions)
      throws IOException {
    writeKey (writer);
    writeLocation (writer, diana_extensions);
    writeQualifiers (writer, diana_extensions);
  }

  /**
   *  Read and return a PublicDBStreamFeature from a stream.  A feature must
   *  be the next thing in the stream.
   *  @param stream the Feature is read from this stream
   *  @param feature_type this flag indicates whether to read the feature as
   *    an EMBL feature (flag == LineGroup.EMBL_FEATURE_TABLE) or as a GENBANK
   *    feature (flag == LineGroup.GENBANK_FEATURE_TABLE).
   *  @exception IOException thrown if there is a problem reading the Feature -
   *    most likely ReadFormatException.
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    the given Qualifier.
   *  @return null if in_stream is at the end of file when the method is called */
  protected static PublicDBStreamFeature readFromStream (LinePushBackReader stream,
                                                         int feature_type)
      throws IOException, InvalidRelationException {

    final PublicDBStreamFeature new_feature;

    if (feature_type == LineGroup.EMBL_FEATURE_TABLE) {
      new_feature = new EmblStreamFeature ();
    } else {
      new_feature = new GenbankStreamFeature ();
    }

    try {
      new_feature.setFromStream (stream);
    } catch (EOFException e) {
      return null;
    }

    return new_feature;
  }

  /**
   *  Return the type of this LineGroup object (SEQUENCE, FEATURE_TABLE, etc.)
   *  Returns FEATURE_TABLE for objects of this class.
   **/
  public int getType () {
    return FEATURE_TABLE;
  }

  /**
   *  This is used by readFromStream () as temporary storage.  It is a class
   *  member rather than a local variable so that we don't need to allocate a
   *  object for each call.  The number we pick for the initial StringBuffer
   *  size is not critical, but should cover most possibilities
   **/
  final static private StringBuffer qualifier_string_buffer =
    new StringBuffer (1500);

  /**
   *  Read the details of a feature from an EMBL stream into the current
   *  object.
   *  @param in_stream the Feature is read from this stream
   *  @exception IOException thrown if there is a problem reading the Feature -
   *    most likely ReadFormatException.
   *  @exception InvalidRelationException Thrown if this Feature cannot contain
   *    the given Qualifier.
   **/
  public void setFromStream (final LinePushBackReader in_stream)
      throws IOException, InvalidRelationException, ReadOnlyException {

    final String first_line = in_stream.readLine ();

    if (first_line == null) {
      // we tried to read a Feature starting at the end of file
      throw new EOFException ("while reading a feature");
    }

    final int line_type = getLineType (first_line);

//      System.err.println ("first_line: " + first_line + " type: " + line_type +
//                          " " + this);

    if (!(this instanceof EmblStreamFeature &&
          line_type == EMBL_FEATURE_TABLE ||
          this instanceof GenbankStreamFeature &&
          line_type == GENBANK_FEATURE_TABLE)) {
      // this line is not the first line of a feature

      //      System.out.println ("in Feature () line type: " + line_type);
      in_stream.pushBack (first_line);

      throw new EOFException ("end of feature table");
    }

    if (first_line.length () < 15) {
      throw new ReadFormatException ("line too short at line: ",
                                     in_stream.getLineNumber ());
    }

    final String key_string =
      getKeyStringFromLine (first_line, in_stream.getLineNumber ());

    if (key_string == null) {
      throw new ReadFormatException ("expected the first line of a " +
                                     "feature at line: ",
                                     in_stream.getLineNumber ());
    }

    // set to true when we see the first qualifier.
    // we need this so we know when to stop adding to location_string and
    // start adding to qualifier_string.
    boolean location_string_finished = false;

    String location_string = getRestOfFeatureLine (first_line);

    qualifier_string_buffer.setLength (0);

    // the number we pick for the initial StringBuffer size is not critical,
    // but should cover most possibilities so that the buffer does not have to
    // increase its size
    qualifier_string_buffer.ensureCapacity (1500);

    String line;

    // the line of the input where the qualifiers start - used for error
    // reporting
    int qualifier_start_line = -1;

    // the line of the input where this feature starts - used for error
    // reporting
    final int feature_start_line = in_stream.getLineNumber ();

    // loop until there are no more lines in the file or we hit the start of
    // the next feature or the start of the next line group
    while (true) {
      line = in_stream.readLine ();

      if (line == null) {
        // we have reached the end of file - break out of loop
        break;
      }

      final int current_line_type = getLineType (line);

//        System.out.println ("in Feature () loop - current_line: " + line + " " +
//                            current_line_type);

      if (this instanceof EmblStreamFeature &&
          current_line_type == EMBL_FEATURE_TABLE ||
          this instanceof GenbankStreamFeature &&
          current_line_type == GENBANK_FEATURE_TABLE) {

        if (getKeyStringFromLine (line, in_stream.getLineNumber ()) == null) {

          // read the text in this line after the key
          final String rest_of_line = getRestOfFeatureLine (line);

          if (rest_of_line == null) {
//              System.out.println ("current_line_type: " + current_line_type +
//                                  " - " +  line);

            throw new ReadFormatException ("line too short while reading " +
                                           "feature",
                                           in_stream.getLineNumber ());
          }

          if (location_string_finished) {

            // we put space in the string to handle those cases when a
            // qualifier wraps over the end of a line
            qualifier_string_buffer.append (" ").append (rest_of_line);
          } else {
            // the last line(s) we read was part of location string

            if (rest_of_line.startsWith ("/")) {

              // we have now seen the first qualifier line so we can use the
              // location_string to create a Location object

              location_string_finished = true;

              qualifier_string_buffer.append (rest_of_line);

              qualifier_start_line = in_stream.getLineNumber ();

            } else {
              location_string = location_string + rest_of_line;
            }
          }
          // continue with loop
        } else {

          // line has a feature key then it must be the start of a new
          // feature
          in_stream.pushBack (line);
          break;

        }
      } else {
        //          System.out.println ("this line is not a feature line so return now");

        // this line is not a feature line so return now
        in_stream.pushBack (line);
        break;
      }
    }


    final Key key;

    try {
      key = new Key (key_string);
    } catch (InvalidKeyException exception) {
      // re-throw the exception with the line number added

      final String new_error_string = exception.getMessage ();

      System.err.println (new_error_string);

      throw new ReadFormatException (new_error_string,
                                     feature_start_line);
    }


    final Location location;

    try {
      location = new Location (location_string);
    } catch (LocationParseException exception) {
      // re-throw the exception with the line number added

      final String new_error_string = exception.getMessage ();

      System.err.println (new_error_string);

      // subtract 1 because the error was on the previous line
      throw new ReadFormatException (new_error_string,
                                     feature_start_line);
    }

    //    System.out.println ("qualifier_string: " +
    //                        qualifier_string_buffer.toString ());


    final QualifierVector qualifiers;

    try {
      final String qualifier_string = qualifier_string_buffer.toString ();
      qualifiers = getQualifiersFromString (qualifier_string);
    } catch (QualifierParseException exception) {
      // re-throw the exception with the line number added

      final String new_error_string = exception.getMessage ();
      // subtract 1 because the error was on the previous line

      throw new ReadFormatException (new_error_string,
                                     qualifier_start_line);
    }

    set (key, location, qualifiers);
  }


  /**
   *  Get the text in a line from a feature minus the FT tag and the key
   *  @param line the entry line to process
   */
  private static String getRestOfFeatureLine (String line) {

    if (line.length () < LOCATION_START_COLUMN) {
      return null;
    } else {
      return line.substring (LOCATION_START_COLUMN).trim ();
    }
  }

  private final static int KEY_FIELD_WIDTH = 16;

  /**
   *  Return the key from a embl entry line.
   *  @param line_string the text of the entry line to process
   *  @param line_number the line number of line_string in the input stream
   *  @return null if this isn't the first line of a feature, otherwise the
   *    key of this feature
   */
  private static String getKeyStringFromLine (final String line_string,
                                              final int line_number)
      throws ReadFormatException {
    // the first line of a feature starts with "FT   " - remove that first
    final String rest_of_line = getRestOfLine (line_string);

    if (rest_of_line == null ||
        rest_of_line.startsWith (" ")) {
      // this isn't the first line of a feature
      return null;
    } else {
      if (rest_of_line.length () < KEY_FIELD_WIDTH) {
        return null;
      } else {
        if (rest_of_line.charAt (KEY_FIELD_WIDTH - 1) != ' ') {
          throw new ReadFormatException ("column " +
                                         LOCATION_START_COLUMN + " must "
                                         + "be empty", line_number);
        }

        final String key_field = rest_of_line.substring (0,15);

        //      System.out.println ("in getKeyStringFromLine () key: " + key_field.trim ());

        return key_field.trim ();
      }
    }
  }

  /**
   *  Read some embl feature qualifiers from a stream into a QualifierVector
   *  object.  The stream should contain qualifiers in this form:
   *  <PRE>  /name1=value1/name2="value2"/name3=[value3]  </PRE>
   *  @param in_stream the qualifiers are read from this stream
   *  @exception IOException thrown if there is a problem reading the
   *    qualifiers, such as end of file.
   *  @exception QualifierParseException Thrown if the format of the value
   *    String is not appropriate for a Qualifier with the given name.  Each
   *    qualifier has a specific format for the value part which depends on
   *    the name, for example the value part of /codon_start qualifier must be
   *    a number: 1, 2 or 3.
   *  @return A Vector containing one Qualifier object for each name/value
   *    pair read from the stream.
   **/
  public static QualifierVector readQualifiers (Reader in_stream)
      throws QualifierParseException, InvalidQualifierException, IOException {

    QualifierVector return_vector = new QualifierVector ();

    BufferedReader buffered_reader = new BufferedReader (in_stream);

    String name;
    String value;

    // loop until end of file
    while (true) {

      name = StreamQualifier.readName (buffered_reader);

      if (name == null) {
        // end of file/stream
        break;
      }

      // save one character in case the next char is not a '='
      buffered_reader.mark (1);

      final int next_char = buffered_reader.read ();

      if (next_char == -1) {
        value = null;
      } else {
        if (next_char == '=') {
          value = StreamQualifier.readValue (buffered_reader);
        } else {
          // this qualifier doesn't have a value
          value = null;
          buffered_reader.reset ();
        }
      }

      final Qualifier new_qualifier;

      if (value == null) {
        new_qualifier = new Qualifier (name);
      } else {
        new_qualifier = StreamQualifier.makeStreamQualifier (name, value);
      }

      return_vector.addQualifierValues (new_qualifier);
    }

    return return_vector;
  }


  /**
   *  Return a QualifierVector containing the qualifiers from a String.
   *  @param qual_string contains the qualifiers to parse
   */
  private static QualifierVector getQualifiersFromString (String qual_string)
      throws QualifierParseException {

    final StringReader string_reader = new StringReader (qual_string);

    final QualifierVector qualifiers;

    try {
      qualifiers = readQualifiers (string_reader);
    } catch (IOException exception) {
      throw (new QualifierParseException (exception.getMessage ()));
    }

    string_reader.close ();

    return qualifiers;
  }

  /**
   *  The column of the output where we should start writting the key.
   **/
  private final static int KEY_START_COLUMN = 5;

  /**
   *  The column of the output where we should start writting the location.
   **/
  private final static int LOCATION_START_COLUMN = 21;

  /**
   *  We should wrap at this column - no characters will be put in or after
   *  this column
   **/
  private final static int LAST_COLUMN = 81;

  /**
   *  Write the key of this Feature to the given stream.
   **/
  private void writeKey (Writer writer)
      throws IOException {
    if (this instanceof EmblStreamFeature) {
      writer.write ("FT   " + getKey ());
    } else {
      writer.write ("     " + getKey ());
    }

    final StringBuffer spaces = new StringBuffer (20);

    // add spaces so that the location starts at coloumn 21
    for (int i = 0 ;
         i < LOCATION_START_COLUMN - KEY_START_COLUMN - getKey ().length () ;
         ++i) {
      spaces.append (' ');
    }

    writer.write (spaces.toString ());
  }


  /**
   *  Write the location of this feature to a stream.  It is written in the
   *  usual EMBL format.  Line that are more than 79 characters wide are
   *  wrapped.  The wrapped lines start with "FT                   ", the
   *  first line doesn't.
   *  @param writer The stream to write to.
   *  @param diana_extensions If true then the any diana additions to the embl
   *    file format will be included in the output, otherwise they will be
   *    removed.  If false for this method then long location strings will be
   *    split over many lines and where the location is a complemented join
   *    it will be written as "join(complement(..." rather than
   *    "complement(join(...".
   *  @exception IOException thrown if there is an io problem while writing
   *    the Feature.
   **/
  private void writeLocation (final Writer writer,
                              final boolean diana_extensions)
      throws IOException {


    if (diana_extensions) {
      String location_string = getLocation ().toStringJoinInner ();

      // write it out on one line
      writer.write (location_string);
      writer.write ("\n");
    } else {
      String location_string = getLocation ().toString ();

      // write the first line without the "FT ..." bit as this will already
      // have been written by writeKey ()

      while (true) {
        final int wrap_position =
          getWrapPosition (location_string,
                           ',',
                           LAST_COLUMN - LOCATION_START_COLUMN - 1);
        if (wrap_position == -1) {
          // write out the remainer of the location_string
          writer.write (location_string);
          writer.write ("\n");
          break;
        } else {
          // write the start of the string (the bit before wrap_position) and
          // save the rest for the next time around the loop

          writer.write (location_string.substring (0, wrap_position));
          if (this instanceof EmblStreamFeature) {
            writer.write ("\n" + "FT                   ");
          } else {
            writer.write ("\n" + "                     ");
          }

          location_string = location_string.substring (wrap_position);
        }
      }
    }
  }

  /**
   *  Write the qualifiers of this feature to a stream.  The qualifiers are
   *  written in EMBL format: ie. <p>
   *  FT                   /codon_start=1
   *  <p> etc.
   *  @param writer The stream to write to.
   *  @param include_diana_extensions If true then the any diana additions to
   *    the embl file format will be included in the output, otherwise they
   *    will be removed.
   *  @exception IOException thrown if there is an io problem while writing
   *    the Feature.
   **/
  private void writeQualifiers (final Writer writer,
                                final boolean include_diana_extensions)
      throws IOException {
    for (int i = 0 ; i < qualifiers.size () ; ++i) {
      final Qualifier current_qualifier = qualifiers.elementAt (i);

      final String qualifier_name = current_qualifier.getName ();

      if (!include_diana_extensions &&
          !EntryInformation.isPossibleEMBLQualifier (qualifier_name) ||
          current_qualifier.getName ().equals ("label") &&
          current_qualifier.getValues ().size () == 1 &&
          current_qualifier.getValues ().elementAt (0).equals ("*")) {
        continue;
      }

      // this will contain one String for each /name=value pair
      final StringVector qualifier_strings =
        StreamQualifier.toStringVector (current_qualifier);

      for (int value_index = 0 ;
           value_index < qualifier_strings.size () ;
           ++value_index) {

        String qualifier_string = qualifier_strings.elementAt (value_index);

        while (true) {
          final int wrap_position =
            getWrapPosition (qualifier_string,
                             ' ',
                             LAST_COLUMN - LOCATION_START_COLUMN);
          if (wrap_position == -1) {
            // write out the remainer of the qualifier_string
            if (this instanceof EmblStreamFeature) {
              writer.write ("FT                   " + qualifier_string);
            } else {
              writer.write ("                     " + qualifier_string);
            }
            writer.write ("\n");
            break;
          } else {
            // write the start of the string (the bit before wrap_position) and
            // save the rest for the next time around the loop

            if (this instanceof EmblStreamFeature) {
              writer.write ("FT                   " +
                            qualifier_string.substring (0, wrap_position - 1));
            } else {
              writer.write ("                     " +
                            qualifier_string.substring (0, wrap_position - 1));
            }
            writer.write ("\n");

            qualifier_string = qualifier_string.substring (wrap_position);
          }
        }
      }
    }
  }

  /**
   *  Find a suitable character position to wrap at.
   *  (Helper method for writeLocation () and writeQualifiers ().)
   *  @param string_to_wrap This is the string to search for the wrap
   *    character.
   *  @param wrap_character This is the character after which we should wrap
   *    the line.
   *  @param line_length The maximum line length before wrapping.
   *  @return The character position in the string_to_wrap that should start
   *    the next line or wrap or -1 if there is no suitable position.
   **/
  private int getWrapPosition (String string_to_wrap,
                               char wrap_character,
                               int line_length) {
    final int return_index;

    if (string_to_wrap.length () < line_length) {
      return -1;                // no need to wrap
    }

    final int wrap_char_index =
      string_to_wrap.substring (0, line_length).lastIndexOf (wrap_character);

    if (wrap_char_index == -1) {
      // there is no wrap_character within the line_length characters - so
      // just return the first occurrence in the string (or -1).
      return string_to_wrap.indexOf (wrap_character);
    } else {
      return wrap_char_index + 1;
    }
  }

  /**
   *  The DocumentEntry object that contains this Feature as passed to the
   *  constructor.
   **/
  private DocumentEntry entry;

  /**
   *  The is the key that was passed to the constructor
   **/
  private Key key;

  /**
   *  The is the Location object that was passed to the constructor
   **/
  private Location location;

  /**
   *  The is the vector of qualifiers that was passed to the constructor
   **/
  private QualifierVector qualifiers = new QualifierVector ();

  /**
   *  The first base of the location of this feature.  This is a cache.
   **/
  private int first_base = -1;

  /**
   *  The last base of the location of this feature.  This is a cache.
   **/
  private int last_base = -1;

  /**
   *  A unique identifier for this feature.
   **/
  private final long id = id_counter++;
}
